@@ -38,7 +38,7 @@ use url::Url;
 
 use crate::aws::client::{CompleteMultipartMode, PutPartPayload, RequestError, S3Client};
 use crate::client::get::GetClientExt;
-use crate::client::list::ListClientExt;
+use crate::client::list::{ListClient, ListClientExt};
 use crate::client::CredentialProvider;
 use crate::multipart::{MultipartStore, PartId};
 use crate::signer::Signer;
@@ -78,6 +78,7 @@ const STORE: &str = "S3";
 /// [`CredentialProvider`] for [`AmazonS3`]
 pub type AwsCredentialProvider = Arc<dyn CredentialProvider<Credential = AwsCredential>>;
 use crate::client::parts::Parts;
+use crate::list::{PaginatedListOptions, PaginatedListResult, PaginatedListStore};
 pub use credential::{AwsAuthorizer, AwsCredential};
 
 /// Interface for [Amazon S3](https://aws.amazon.com/s3/).
@@ -196,6 +197,7 @@ impl ObjectStore for AmazonS3 {
                     r => r,
                 }
             }
+            #[allow(deprecated)]
             (PutMode::Create, S3ConditionalPut::Dynamo(d)) => {
                 d.conditional_op(&self.client, location, None, move || request.do_put())
                     .await
@@ -227,6 +229,7 @@ impl ObjectStore for AmazonS3 {
                             r => r,
                         }
                     }
+                    #[allow(deprecated)]
                     S3ConditionalPut::Dynamo(d) => {
                         d.conditional_op(&self.client, location, Some(&etag), move || {
                             request.do_put()
@@ -366,6 +369,7 @@ impl ObjectStore for AmazonS3 {
 
                 return res;
             }
+            #[allow(deprecated)]
             Some(S3CopyIfNotExists::Dynamo(lock)) => {
                 return lock.copy_if_not_exists(&self.client, from, to).await
             }
@@ -493,10 +497,23 @@ impl MultipartStore for AmazonS3 {
     }
 }
 
+#[async_trait]
+impl PaginatedListStore for AmazonS3 {
+    async fn list_paginated(
+        &self,
+        prefix: Option<&str>,
+        opts: PaginatedListOptions,
+    ) -> Result<PaginatedListResult> {
+        self.client.list_request(prefix, opts).await
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
     use crate::client::get::GetClient;
+    use crate::client::retry::RetryContext;
+    use crate::client::SpawnedReqwestConnector;
     use crate::integration::*;
     use crate::tests::*;
     use crate::ClientOptions;
@@ -510,7 +527,9 @@ mod tests {
     async fn write_multipart_file_with_signature() {
         maybe_skip_integration!();
 
+        let bucket = "test-bucket-for-checksum";
         let store = AmazonS3Builder::from_env()
+            .with_bucket_name(bucket)
             .with_checksum_algorithm(Checksum::SHA256)
             .build()
             .unwrap();
@@ -588,6 +607,7 @@ mod tests {
         signing(&integration).await;
         s3_encryption(&integration).await;
         put_get_attributes(&integration).await;
+        list_paginated(&integration, &integration).await;
 
         // Object tagging is not supported by S3 Express One Zone
         if config.session_provider.is_none() {
@@ -622,6 +642,7 @@ mod tests {
         put_get_delete_list(&integration).await;
 
         match &integration.client.config.copy_if_not_exists {
+            #[allow(deprecated)]
             Some(S3CopyIfNotExists::Dynamo(d)) => dynamo::integration_test(&integration, d).await,
             _ => eprintln!("Skipping dynamo integration test - dynamo not configured"),
         };
@@ -741,11 +762,14 @@ mod tests {
         upload.complete().await.unwrap();
 
         for location in &locations {
+            let mut context = RetryContext::new(&store.client.config.retry_config);
+
             let res = store
                 .client
-                .get_request(location, GetOptions::default())
+                .get_request(&mut context, location, GetOptions::default())
                 .await
                 .unwrap();
+
             let headers = res.headers();
             assert_eq!(
                 headers
@@ -797,11 +821,14 @@ mod tests {
 
         // Test get with sse-c.
         for location in &locations {
+            let mut context = RetryContext::new(&store.client.config.retry_config);
+
             let res = store
                 .client
-                .get_request(location, GetOptions::default())
+                .get_request(&mut context, location, GetOptions::default())
                 .await
                 .unwrap();
+
             let headers = res.headers();
             assert_eq!(
                 headers
@@ -820,4 +847,57 @@ mod tests {
             store.delete(location).await.unwrap();
         }
     }
+
+    /// Integration test that ensures I/O is done on an alternate threadpool
+    /// when using the `SpawnedReqwestConnector`.
+    #[test]
+    fn s3_alternate_threadpool_spawned_request_connector() {
+        maybe_skip_integration!();
+        let (shutdown_tx, shutdown_rx) = tokio::sync::oneshot::channel::<()>();
+
+        // Runtime with I/O enabled
+        let io_runtime = tokio::runtime::Builder::new_current_thread()
+            .enable_all() // <-- turns on IO
+            .build()
+            .unwrap();
+
+        // Runtime without I/O enabled
+        let non_io_runtime = tokio::runtime::Builder::new_current_thread()
+            // note: no call to enable_all
+            .build()
+            .unwrap();
+
+        // run the io runtime in a different thread
+        let io_handle = io_runtime.handle().clone();
+        let thread_handle = std::thread::spawn(move || {
+            io_runtime.block_on(async move {
+                shutdown_rx.await.unwrap();
+            });
+        });
+
+        let store = AmazonS3Builder::from_env()
+            // use different bucket to avoid collisions with other tests
+            .with_bucket_name("test-bucket-for-spawn")
+            .with_http_connector(SpawnedReqwestConnector::new(io_handle))
+            .build()
+            .unwrap();
+
+        // run a request on the non io runtime -- will fail if the connector
+        // does not spawn the request to the io runtime
+        non_io_runtime
+            .block_on(async move {
+                let path = Path::from("alternate_threadpool/test.txt");
+                store.delete(&path).await.ok(); // remove the file if it exists from prior runs
+                store.put(&path, "foo".into()).await?;
+                let res = store.get(&path).await?.bytes().await?;
+                assert_eq!(res.as_ref(), b"foo");
+                store.delete(&path).await?; // cleanup
+                Ok(()) as Result<()>
+            })
+            .expect("failed to run request on non io runtime");
+
+        // shutdown the io runtime and thread
+        shutdown_tx.send(()).ok();
+        thread_handle.join().expect("runtime thread panicked");
+    }
 }
