@@ -19,19 +19,21 @@ use crate::client::builder::HttpRequestBuilder;
 use crate::client::get::GetClient;
 use crate::client::header::{get_put_result, get_version, HeaderConfig};
 use crate::client::list::ListClient;
-use crate::client::retry::RetryExt;
+use crate::client::retry::{RetryContext, RetryExt};
 use crate::client::s3::{
     CompleteMultipartUpload, CompleteMultipartUploadResult, InitiateMultipartUploadResult,
     ListResponse,
 };
 use crate::client::{GetOptionsExt, HttpClient, HttpError, HttpResponse};
+use crate::gcp::credential::CredentialExt;
 use crate::gcp::{GcpCredential, GcpCredentialProvider, GcpSigningCredentialProvider, STORE};
+use crate::list::{PaginatedListOptions, PaginatedListResult};
 use crate::multipart::PartId;
-use crate::path::{Path, DELIMITER};
+use crate::path::Path;
 use crate::util::hex_encode;
 use crate::{
-    Attribute, Attributes, ClientOptions, GetOptions, ListResult, MultipartId, PutMode,
-    PutMultipartOpts, PutOptions, PutPayload, PutResult, Result, RetryConfig,
+    Attribute, Attributes, ClientOptions, GetOptions, MultipartId, PutMode, PutMultipartOpts,
+    PutOptions, PutPayload, PutResult, Result, RetryConfig,
 };
 use async_trait::async_trait;
 use base64::prelude::BASE64_STANDARD;
@@ -144,30 +146,21 @@ pub(crate) struct GoogleCloudStorageConfig {
     pub retry_config: RetryConfig,
 
     pub client_options: ClientOptions,
+
+    pub skip_signature: bool,
 }
 
 impl GoogleCloudStorageConfig {
-    pub(crate) fn new(
-        base_url: String,
-        credentials: GcpCredentialProvider,
-        signing_credentials: GcpSigningCredentialProvider,
-        bucket_name: String,
-        retry_config: RetryConfig,
-        client_options: ClientOptions,
-    ) -> Self {
-        Self {
-            base_url,
-            credentials,
-            signing_credentials,
-            bucket_name,
-            retry_config,
-            client_options,
-        }
-    }
-
     pub(crate) fn path_url(&self, path: &Path) -> String {
         format!("{}/{}/{}", self.base_url, self.bucket_name, path)
     }
+
+    pub(crate) async fn get_credential(&self) -> Result<Option<Arc<GcpCredential>>> {
+        Ok(match self.skip_signature {
+            false => Some(self.credentials.get_credential().await?),
+            true => None,
+        })
+    }
 }
 
 /// A builder for a put request allowing customisation of the headers and query string
@@ -304,8 +297,8 @@ impl GoogleCloudStorageClient {
         &self.config
     }
 
-    async fn get_credential(&self) -> Result<Arc<GcpCredential>> {
-        self.config.credentials.get_credential().await
+    async fn get_credential(&self) -> Result<Option<Arc<GcpCredential>>> {
+        self.config.get_credential().await
     }
 
     /// Create a signature from a string-to-sign using Google Cloud signBlob method.
@@ -341,7 +334,7 @@ impl GoogleCloudStorageClient {
         let response = self
             .client
             .post(&url)
-            .bearer_auth(&credential.bearer)
+            .with_bearer_auth(credential.as_deref())
             .json(&body)
             .retryable(&self.config.retry_config)
             .idempotent(true)
@@ -493,7 +486,7 @@ impl GoogleCloudStorageClient {
 
         self.client
             .request(Method::DELETE, &url)
-            .bearer_auth(&credential.bearer)
+            .with_bearer_auth(credential.as_deref())
             .header(CONTENT_TYPE, "application/octet-stream")
             .header(CONTENT_LENGTH, "0")
             .query(&[("uploadId", multipart_id)])
@@ -514,13 +507,11 @@ impl GoogleCloudStorageClient {
         completed_parts: Vec<PartId>,
     ) -> Result<PutResult> {
         if completed_parts.is_empty() {
-            // GCS doesn't allow empty multipart uploads
-            let result = self
-                .request(Method::PUT, path)
-                .idempotent(true)
-                .do_put()
-                .await?;
+            // GCS doesn't allow empty multipart uploads, so fallback to regular upload.
             self.multipart_cleanup(path, multipart_id).await?;
+            let result = self
+                .put(path, PutPayload::new(), Default::default())
+                .await?;
             return Ok(result);
         }
 
@@ -540,7 +531,7 @@ impl GoogleCloudStorageClient {
         let response = self
             .client
             .request(Method::POST, &url)
-            .bearer_auth(&credential.bearer)
+            .with_bearer_auth(credential.as_deref())
             .query(&[("uploadId", upload_id)])
             .body(data)
             .retryable(&self.config.retry_config)
@@ -596,7 +587,7 @@ impl GoogleCloudStorageClient {
         }
 
         builder
-            .bearer_auth(&credential.bearer)
+            .with_bearer_auth(credential.as_deref())
             // Needed if reqwest is compiled with native-tls instead of rustls-tls
             // See https://github.com/apache/arrow-rs/pull/3921
             .header(CONTENT_LENGTH, 0)
@@ -626,8 +617,17 @@ impl GetClient for GoogleCloudStorageClient {
         user_defined_metadata_prefix: Some(USER_DEFINED_METADATA_HEADER_PREFIX),
     };
 
+    fn retry_config(&self) -> &RetryConfig {
+        &self.config.retry_config
+    }
+
     /// Perform a get request <https://cloud.google.com/storage/docs/xml-api/get-object-download>
-    async fn get_request(&self, path: &Path, options: GetOptions) -> Result<HttpResponse> {
+    async fn get_request(
+        &self,
+        ctx: &mut RetryContext,
+        path: &Path,
+        options: GetOptions,
+    ) -> Result<HttpResponse> {
         let credential = self.get_credential().await?;
         let url = self.object_url(path);
 
@@ -642,13 +642,11 @@ impl GetClient for GoogleCloudStorageClient {
             request = request.query(&[("generation", version)]);
         }
 
-        if !credential.bearer.is_empty() {
-            request = request.bearer_auth(&credential.bearer);
-        }
-
         let response = request
+            .with_bearer_auth(credential.as_deref())
             .with_get_options(options)
-            .send_retry(&self.config.retry_config)
+            .retryable_request()
+            .send(ctx)
             .await
             .map_err(|source| {
                 let path = path.as_ref().into();
@@ -665,24 +663,22 @@ impl ListClient for Arc<GoogleCloudStorageClient> {
     async fn list_request(
         &self,
         prefix: Option<&str>,
-        delimiter: bool,
-        page_token: Option<&str>,
-        offset: Option<&str>,
-    ) -> Result<(ListResult, Option<String>)> {
+        opts: PaginatedListOptions,
+    ) -> Result<PaginatedListResult> {
         let credential = self.get_credential().await?;
         let url = format!("{}/{}", self.config.base_url, self.bucket_name_encoded);
 
         let mut query = Vec::with_capacity(5);
         query.push(("list-type", "2"));
-        if delimiter {
-            query.push(("delimiter", DELIMITER))
+        if let Some(delimiter) = &opts.delimiter {
+            query.push(("delimiter", delimiter.as_ref()))
         }
 
-        if let Some(prefix) = &prefix {
+        if let Some(prefix) = prefix {
             query.push(("prefix", prefix))
         }
 
-        if let Some(page_token) = page_token {
+        if let Some(page_token) = &opts.page_token {
             query.push(("continuation-token", page_token))
         }
 
@@ -690,15 +686,22 @@ impl ListClient for Arc<GoogleCloudStorageClient> {
             query.push(("max-keys", max_results))
         }
 
-        if let Some(offset) = offset {
-            query.push(("start-after", offset))
+        if let Some(offset) = &opts.offset {
+            query.push(("start-after", offset.as_ref()))
+        }
+
+        let max_keys_str;
+        if let Some(max_keys) = &opts.max_keys {
+            max_keys_str = max_keys.to_string();
+            query.push(("max-keys", max_keys_str.as_ref()))
         }
 
         let response = self
             .client
             .request(Method::GET, url)
+            .extensions(opts.extensions)
             .query(&query)
-            .bearer_auth(&credential.bearer)
+            .with_bearer_auth(credential.as_deref())
             .send_retry(&self.config.retry_config)
             .await
             .map_err(|source| Error::ListRequest { source })?
@@ -711,6 +714,9 @@ impl ListClient for Arc<GoogleCloudStorageClient> {
             .map_err(|source| Error::InvalidListResponse { source })?;
 
         let token = response.next_continuation_token.take();
-        Ok((response.try_into()?, token))
+        Ok(PaginatedListResult {
+            result: response.try_into()?,
+            page_token: token,
+        })
     }
 }
