@@ -15,6 +15,7 @@
 // specific language governing permissions and limitations
 // under the License.
 
+#![cfg_attr(docsrs, feature(doc_auto_cfg))]
 #![deny(rustdoc::broken_intra_doc_links, rustdoc::bare_urls, rust_2018_idioms)]
 #![warn(
     missing_copy_implementations,
@@ -496,6 +497,14 @@
 //! [`rustls-native-certs`]: https://crates.io/crates/rustls-native-certs/
 //! [`webpki-roots`]: https://crates.io/crates/webpki-roots
 //!
+//! # Customizing HTTP Clients
+//!
+//! Many [`ObjectStore`] implementations permit customization of the HTTP client via
+//! the [`HttpConnector`] trait and utilities in the [`client`] module.
+//! Examples include injecting custom HTTP headers or using an alternate
+//! tokio Runtime I/O requests.
+//!
+//! [`HttpConnector`]: client::HttpConnector
 
 #[cfg(feature = "aws")]
 pub mod aws;
@@ -515,6 +524,7 @@ pub mod local;
 pub mod memory;
 pub mod path;
 pub mod prefix;
+pub mod registry;
 #[cfg(feature = "cloud")]
 pub mod signer;
 pub mod throttle;
@@ -538,6 +548,7 @@ mod tags;
 
 pub use tags::TagSet;
 
+pub mod list;
 pub mod multipart;
 mod parse;
 mod payload;
@@ -600,6 +611,8 @@ pub trait ObjectStore: std::fmt::Display + Send + Sync + Debug + 'static {
     ///
     /// Client should prefer [`ObjectStore::put`] for small payloads, as streaming uploads
     /// typically require multiple separate requests. See [`MultipartUpload`] for more information
+    ///
+    /// For more advanced multipart uploads see [`MultipartStore`](multipart::MultipartStore)
     async fn put_multipart(&self, location: &Path) -> Result<Box<dyn MultipartUpload>> {
         self.put_multipart_opts(location, PutMultipartOpts::default())
             .await
@@ -609,6 +622,8 @@ pub trait ObjectStore: std::fmt::Display + Send + Sync + Debug + 'static {
     ///
     /// Client should prefer [`ObjectStore::put`] for small payloads, as streaming uploads
     /// typically require multiple separate requests. See [`MultipartUpload`] for more information
+    ///
+    /// For more advanced multipart uploads see [`MultipartStore`](multipart::MultipartStore)
     async fn put_multipart_opts(
         &self,
         location: &Path,
@@ -719,6 +734,8 @@ pub trait ObjectStore: std::fmt::Display + Send + Sync + Debug + 'static {
     /// `foo/bar_baz/x`. List is recursive, i.e. `foo/bar/more/x` will be included.
     ///
     /// Note: the order of returned [`ObjectMeta`] is not guaranteed
+    ///
+    /// For more advanced listing see [`PaginatedListStore`](list::PaginatedListStore)
     fn list(&self, prefix: Option<&Path>) -> BoxStream<'static, Result<ObjectMeta>>;
 
     /// List all the objects with the given prefix and a location greater than `offset`
@@ -727,6 +744,8 @@ pub trait ObjectStore: std::fmt::Display + Send + Sync + Debug + 'static {
     /// the number of network requests required
     ///
     /// Note: the order of returned [`ObjectMeta`] is not guaranteed
+    ///
+    /// For more advanced listing see [`PaginatedListStore`](list::PaginatedListStore)
     fn list_with_offset(
         &self,
         prefix: Option<&Path>,
@@ -975,7 +994,7 @@ impl GetOptions {
     /// Returns an error if the modification conditions on this request are not satisfied
     ///
     /// <https://datatracker.ietf.org/doc/html/rfc7232#section-6>
-    fn check_preconditions(&self, meta: &ObjectMeta) -> Result<()> {
+    pub fn check_preconditions(&self, meta: &ObjectMeta) -> Result<()> {
         // The use of the invalid etag "*" means no ETag is equivalent to never matching
         let etag = meta.e_tag.as_deref().unwrap_or("*");
         let last_modified = meta.last_modified;
